/*
 * idiscover.c
 * Discover all IPMI-LAN-enabled nodes on this network or subnet.
 * This program is not completely reliable yet, not all IPMI-LAN-enabled
 * nodes respond.
 *
 * Author:  Andy Cress  arcress@users.sourceforge.net
 * Copyright (c) 2006 Intel Corporation. 
 *
 * 10/27/06 Andy Cress - created
 */
/*M*
Copyright (c) 2006, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include "getopt.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netdb.h>
#include <fcntl.h>
#include <getopt.h>
#include <errno.h>
#include <pthread.h>
#endif
#include <string.h>
#include "ipmicmd.h"
 
#ifndef INET_ADDRSTRLEN
#define INET_ADDRSTRLEN 16
#endif
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

#define NO_THREADS  1
#define RMCP_PRI_RMCP_PORT     0x26F
#define SZ_PING    12
#define IPMI_PING_MAX_LEN  50  /* usu 28 */

/*
 * Global variables 
 */
static char * progver   = "0.1";
static char * progname  = "idiscover";
static char   fdebug    = 0;
static uchar ipmi_maj = 0;
static uchar ipmi_min = 0;
static uchar bSolVer  = 0;
static uchar actcmd;
static uchar netfn;
static int g_sockfd = 0; 
static struct sockaddr_in _srcaddr;
// static struct sockaddr_in _destaddrlist[255];
static struct  in_addr    _startAddr, _endAddr; 
static char g_startDest[MAXHOSTNAMELEN+1];  
static char g_endDest[MAXHOSTNAMELEN+1];
static char g_interface[INET_ADDRSTRLEN+1] = {"eth0"};
static int  g_num_packets = 0;
static int  g_npings = 0;
static int  g_npongs = 0;
static int  g_recv_status = 0;
static int  g_wait   = 1; /* num sec to wait */
static int  g_delay  = 0; /* num usec between sends */

void printerr( const char *pattn, ...)
{
   va_list arglist;
   FILE *fderr;

   // fderr = fopen("/tmp/idiscover.log","a+");
   // if (fderr == NULL) return;
   fderr = stderr;

   va_start(arglist, pattn);
   vfprintf(fderr, pattn, arglist);
   va_end(arglist);

   // fclose(fderr);
}

static void cleanup(void)
{
   if (g_sockfd > 0) close(g_sockfd);
}

void show_usage(void)
{
   printf("Usage: %s [-aisx] -b <begin IP> -e <end IP>\n", progname);
   // printf(" where -a      all node broadcast\n");
   printf("       -i      interface name, default is eth0\n");
   // printf("       -s      specific subnet\n");
   printf("       -x      show eXtra debug messages\n");
}

static int  _sleep(unsigned int s, unsigned int u)
{
  struct timeval tv;
  tv.tv_sec  = s;
  tv.tv_usec = u;
  if (select(1, NULL, NULL, NULL, &tv) < 0)
     printerr("select: %s", strerror(errno));
  return 0;
}

void split_ip(int addr, uchar *ip)
{
   ip[3] = (addr & 0x000000ff);
   ip[2] = (addr & 0x0000ff00) >> 8;
   ip[1] = (addr & 0x00ff0000) >> 16;
   ip[0] = (addr & 0xff000000) >> 24;
}

int ntoi(int addr)
{
   return(addr);
}

void show_ip(int saddr)
{
   uchar ip[4];
   split_ip(saddr,ip);
   printerr("%d.%d.%d.%d\n",ip[0],ip[1],ip[2],ip[3]);
}

static int ipmilan_sendto(int s, const void *msg, size_t len, int flags, 
				const struct sockaddr *to, socklen_t tolen)
{
    int n;
#ifdef NEED_PAD
    int fusepad = 0;
    /* Never need a pad byte for ping/pong packets */
    if (len == 56 || len == 84 || len == 112 || len == 128 || len == 156) {
        fusepad = 1;
        len += 1;
    }
#endif
    n = sendto(s,msg,len,flags,to,tolen);
    // if (fusepad && (n > 0)) n--;
    return(n);
}

static int ipmilan_recvfrom(int s, void *buf, size_t len, int flags, 
		struct sockaddr *from, socklen_t *fromlen)
{
    int rv;
    rv = recvfrom(s,buf,len,flags,from,fromlen);
#ifndef WIN32
    /* Sometimes the Linux OS sends an ECONNREFUSED error, but 
     * retrying will catch the BMC's reply packet. */
    if ((rv < 0) && (errno == ECONNREFUSED))
        rv = recvfrom(s,buf,len,flags,from,fromlen);
#endif
    return(rv);
}

int sock_init( char *_interface, char *_startIP, char *_endIP)
{
  struct hostent *hptr;
  int rv, i;
 
  if ((g_sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    printerr("socket: %s", strerror(errno));
    return(-1);
  }

  memset(&_srcaddr, 0, sizeof(_srcaddr));
  _srcaddr.sin_family = AF_INET;
  _srcaddr.sin_port = htons(0);
 
  if (_interface == NULL)
    _srcaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  else
   {
      if (strchr(_interface, '.') != NULL)
        {   /* assume it is an IP address*/
          if ((rv = inet_pton(AF_INET, _interface, &_srcaddr.sin_addr)) < 0)
            printerr("inet_pton: %s", strerror(errno));
          if (rv == 0) 
            printerr("invalid interface address");
          return(rv);
        }
      else
        {   /* assume interface name, like eth0 */
          struct ifreq ifr;
          struct sockaddr_in temp_sockaddr;
 
          strncpy(ifr.ifr_name, _interface, IFNAMSIZ);
          ifr.ifr_addr.sa_family = AF_INET;
          if (ioctl(g_sockfd, SIOCGIFADDR, &ifr) < 0) {
             printerr("ioctl: %s", strerror(errno));
             return(-1);
          }
 
          temp_sockaddr = *((struct sockaddr_in *)&ifr.ifr_addr);
          memcpy(&_srcaddr.sin_addr.s_addr, &temp_sockaddr.sin_addr.s_addr,
                 sizeof(_srcaddr.sin_addr.s_addr));
        }
   }
 
   _srcaddr.sin_addr.s_addr = htonl(INADDR_ANY);
   if (bind(g_sockfd, (struct sockaddr *)&_srcaddr, sizeof(_srcaddr)) < 0) {
      printerr("bind: %s", strerror(errno));
      return(-1);
   }
 
   rv = inet_aton(_startIP, &_startAddr);
   if (rv ) {
        _startAddr.s_addr = ntohl(_startAddr.s_addr);
         if (fdebug) show_ip(_startAddr.s_addr);
         if ((*((unsigned char*)&_startAddr.s_addr) <1 )||(*((unsigned char*)&_startAddr.s_addr) >254 ))
                show_usage();
         else rv = 0;
   } else {
        show_usage();
   }
 
   rv = inet_aton(_endIP, &_endAddr);
   if (rv ) {
        _endAddr.s_addr = ntohl(_endAddr.s_addr);
        if (fdebug) show_ip(_endAddr.s_addr);
 
         if ((*((unsigned char*)&_endAddr.s_addr) <1 )||(*((unsigned char*)&_endAddr.s_addr) >254 ))
                show_usage();
         else rv = 0;
   } else {
         show_usage();
   }

   /* calculate g_num_packets */
   g_num_packets = ntoi(_endAddr.s_addr) - ntoi(_startAddr.s_addr) + 1;
   if (fdebug) printerr("g_num_packets = %d\n",g_num_packets);
   if (g_num_packets < 1) g_num_packets = 0;

   return(rv);
}  /*end init*/

void *receiveThread(void *p)
{
   uchar buffer[IPMI_PING_MAX_LEN];
   struct timeval tv;
   fd_set rset;
   int rv, len;

   do
   {
      tv.tv_sec  = g_wait;
      tv.tv_usec = 0;
      FD_ZERO(&rset);
      FD_SET(g_sockfd, &rset);
      g_recv_status = 0;

      if (fdebug) printerr("waiting for ping %d response\n",g_npings);
      if ((rv = select(g_sockfd+1, &rset, NULL, NULL, &tv)) < 0) {
          printerr("select: %s", strerror(errno));
          cleanup();
          exit(-1);
      }
      
      if (fdebug) printerr("select rv = %d\n",rv);
      if (rv >= 1)
          {
              struct sockaddr_in from;
              socklen_t fromlen;
 
              g_recv_status = 1;
              fromlen = sizeof(from);
              len = ipmilan_recvfrom(g_sockfd, buffer, IPMI_PING_MAX_LEN, 0,
                                     (struct sockaddr *)&from, &fromlen);
              if (fdebug) printerr("recvfrom rv = %d\n",rv);
              if (len < 0) { 
		printerr("ipmilan_recvfrom: %s", strerror(errno));
		continue; 
              }
              g_recv_status = 2;
              g_npongs++;
              // parse received pong 
              printf("%d: response from %s\n",g_npongs,
			inet_ntoa(from.sin_addr));
          }

#ifdef NO_THREADS
   } while(0);
#else
   } while(1);
#endif
} 

/*
 * send_ping:
 * RMCP Ping buffer, sent as a UDP packet to port 0x026f.
 * rmcp.ver     = 0x06          // RMCP Version 1.0
 * rmcp.resvd   = 0x00          // RESERVED
 * rmcp.seq     = 0xff          // no RMCP ACK 
 * rmcp.class   = 0x06          // RMCP_CLASS_ASF
 * asf.iana     = 0x000011BE    // ASF_RMCP_IANA (hi-lo)
 * asf.type     = 0x80          // ASF_TYPE_PING
 * asf.tag      = 0x00          // ASF sequence number
 * asf.resvd    = 0x00          // RESERVED
 * asf.len      = 0x00
 */
int send_ping(struct sockaddr_in *_daddr, uchar seq)
{
   uchar pingbuf[SZ_PING] = {06,0,0xFF,06,0x00,0x00,0x11,0xBE,0x80,0,0,0 };
   int rv, len;

   pingbuf[9] = seq;
   len = sizeof(pingbuf);
   rv = ipmilan_sendto(g_sockfd, pingbuf, len, 0, 
	          (struct sockaddr *)_daddr, sizeof(struct sockaddr_in));
   return(rv);
}

static int send_poke1(struct sockaddr_in *_daddr)
{
   int rv;
   uchar asfpkt[16] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                        0x07, 0x20, 0x18, 0xc8, 0xc2, 0x01, 0x01, 0x3c };
   rv = ipmilan_sendto(g_sockfd, asfpkt, 16, 0, 
	          (struct sockaddr *)_daddr, sizeof(struct sockaddr_in));
   return rv;
}

void *sendThread(void *p)
{
   int i, n, len;
   char _dest[MAXHOSTNAMELEN+1];
   char _dest_ip[INET_ADDRSTRLEN+1];
   struct sockaddr_in _destaddr;
   uchar o[4];
   uint ip;
   uchar _seq;
   int rv;

   n = g_num_packets;  /*num*/
   ip = _startAddr.s_addr;

   for (i = 0; i < n; i++) 
   {
      split_ip(ip,o);
      if (o[3] == 0) continue;
      if (o[3] == 255) continue;
      sprintf(_dest_ip,"%d.%d.%d.%d",o[0],o[1],o[2],o[3]);

      /* set _destaddr */
      _destaddr.sin_family = AF_INET;
      _destaddr.sin_port = htons(RMCP_PRI_RMCP_PORT);
      if ( !inet_aton( _dest_ip, &_destaddr.sin_addr)) {
          printerr("inet_aton error %s\n",_dest_ip);
          continue;
      }

      /* send ping buffer */
      _seq = 0;
      rv = send_ping(&_destaddr,_seq);
      g_npings++;
      if (fdebug) printerr("sendto[%d,%d] %s rv = %d\n",
				g_npings,_seq,_dest_ip,rv);
      if (rv < 0) {
          rv = send_ping(&_destaddr,++_seq);
          if (rv < 0) {
              printerr("sendto error %s\n",strerror(errno));
              continue;
          }
      }

#ifdef NO_THREADS
      receiveThread(NULL);
      if (g_recv_status == 0) {
          /* nothing returned, try again */
          rv = send_poke1(&_destaddr);
          if (fdebug) printerr("sendto[%d,%d] %s poke rv = %d\n",
				   g_npings,_seq,_dest_ip,rv);
          rv = send_ping(&_destaddr,++_seq);
          if (fdebug) printerr("sendto[%d,%d] %s rv = %d\n",
				g_npings,_seq,_dest_ip,rv);
          if (rv >= 0) {
              receiveThread(NULL);
          }
      }
#endif

      /* sleep an interval (g_delay usec) */
      if (g_delay > 0) _sleep(0,g_delay);

      ip++; /* increment to next IP */
   }
}

#ifdef NOT
/* Discovery GetDeviceID method, scan slave addresses on a server. */
/* TODO: Should implement a broadcast method for IPMI LAN servers. */
#define NRSSA   10
#define NUM_SA  3
uchar salist[NRSSA];
int nrssa = NRSSA;
int nret = 0;
uchar sa_scan[NUM_SA] = { 0x20, 0x28, 0xC0 };

int Discovery(uchar *list, int max, int *numret)
{
   uchar devrec[16];
   int i, j, ret, sresp;
   uchar sa, cc;
   uchar ipmi_maj, ipmi_min;

   j = 0;
   if (fdebug) printf("Discovery: ");
   for (i = 0; i < NUM_SA; i++) {
	sa = sa_scan[i];
	sresp = sizeof(devrec);
	memset(resp,0,sresp);
	ret = ipmi_cmdraw(GET_DEVICE_ID,NETFN_STOR,sa, 0/*bus*/, 0/*BMC_LUN*/,
		   NULL,0,&devrec[0], &sresp,&cc,fdebug);
	if (ret == 0 && cc == 0) {
	   ipmi_maj = devrec[4] & 0x0f;
	   ipmi_min = devrec[4] >> 4;
	   if (fdebug) printf("%02x ",sa);
	   salist[j++] = sa;
	}
   }
   if (fdebug) printf("\n     numsa = %d",j);
   *numret = j;
   if (j > 0) return(0);
   else return(-1);
}
#endif


#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
{
   int rv = 0;
   uchar devrec[16];
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;
   char c;
   int i;
#ifndef NO_THREADS
   char message[32];
   pthread_t thread[2];
   int  iret[2];
#endif

   printf("%s ver %s\n", progname,progver);

   strcpy(g_startDest,"10.243.42.216");    /*++++ for testing +++++*/
   strcpy(g_endDest,"10.243.42.229");    /*++++*/

   while ( (c = getopt( argc, argv,"ab:e:i:s:x?")) != EOF ) 
      switch(c) {
          case 'a': break;  /*all (broadcast ping)*/
          case 'b':   /*begin*/
                strncpy(g_startDest,optarg,MAXHOSTNAMELEN);
		break;
          case 'e':   /*end*/
                strncpy(g_endDest,optarg,MAXHOSTNAMELEN);
		break;
          case 'i':   /*interface*/
                strncpy(g_interface,optarg,sizeof(g_interface));
		break;
          case 's':   /*subnet*/
                /* copy optarg from 10.243.42.0 or similar, to
                 * begin/end range. */
		break;
          case 'x': fdebug = 1;     break;  /* debug messages */
	  default:
		if (fdebug) printerr("getopt(%c) default\n",c);
                show_usage();
                exit(1);
      }

   rv = sock_init(g_interface, g_startDest, g_endDest); 
   if (fdebug) printerr("init rv = %d, sockfd = %d\n",rv,g_sockfd);
   if (rv != 0) exit(rv);

   printerr("Discovering IPMI Devices:\n");
#ifdef NO_THREADS
   sendThread(NULL);
#else
   iret[0] = pthread_create( &thread[0], NULL, sendThread, (void*) message);
   pthread_join( thread[0], NULL);
   iret[1] = pthread_create( &thread[1], NULL, receiveThread, (void*) message);
   pthread_join( thread[1], NULL);
#endif

   if (fdebug) printerr("%d pings sent, %d responses\n",g_npings,g_npongs);

   cleanup();
   exit (rv);
}  /* end main()*/

/* end idiscover.c */
