/*M*
//  PVCS:
//      $Workfile:   ipmicmd.c  $
//      $Revision:   1.12  $
//      $Modtime:   23 Feb 2005 11:24:14  $
//      $Author:   arcress@users.sourceforge.net  $  
//
//  Define the ipmi_cmd routine and supporting logic to execute IPMI 
//  commands via one of the supported IPMI drivers:
//     /dev/ipmi0  /dev/ipmi/0    = MontaVista OpenIPMI driver
//     /dev/imb                   = Intel IMB ipmidrvr (comes with ISM)
//     /dev/ipmikcs /dev/ipmi/kcs = valinux driver by San Mehat
//     libfreeipmi.so             = GNU FreeIPMI user-space library
//     ldipmidaemon               = LanDesk IPMI daemon (user-space process) 
// 
//  08/05/02 ARC - created 
//  08/15/02 ARC - added decode_cc
//  10/24/02 ARC - made cmd param ushort to be more unique
//  01/29/03 ARC - added MontaVista OpenIPMI driver support
//  07/25/03 ARC - added serial-over-lan commands
//  07/30/03 ARC - added GetThresholds, fix for ipmi_cmd_raw,
//                 changed some error messages
//  09/04/03 ARC - added debug messages for fDriverTyp first time
//  05/05/04 ARC - leave _mv device open, rely on each app calling ipmi_close,
//		   helps performance.
//  08/10/04 ARC - fix typo in ipmi_cmd_raw/mv: cmd->icmd (thanks Kevin Gao)
//  08/26/04 ARC - fix out-of-bounds error in decode_cc
//  10/27/04 ARC - added gnu FreeIPMI library support
//  11/11/04 ARC - added fdebug to ipmi_getdeviceid & ipmi_open_gnu 
//  02/23/05 ARC - added routines for LanDesk, fDriverTyp=5
//  07/15/05 ARC - test for ldipmi first, since it hangs KCS if another 
//                 driver tries to coexist.
//  07/06/06 ARC - better separate driver implementations, cleaner now
 *M*/
/*----------------------------------------------------------------------*
The BSD License 

Copyright (c) 2002-2006, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *----------------------------------------------------------------------*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <termios.h>
#endif

#include "ipmicmd.h"    /* has NCMDS, ipmi_cmd_t */

ipmi_cmd_t ipmi_cmds[NCMDS] = { 
 /*empty,temp*/ 0, BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0, 20, 
 GET_SEL_INFO,     BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0, 14,
 GET_SEL_ENTRY,    BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6, 18,
 RESERVE_SEL,      BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0, 2,
 CLEAR_SEL,        BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6, 1,
 GET_SEL_TIME,     BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6, 4,
 GET_LAN_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4, 19,
 SET_LAN_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 21, 0,
 GET_SER_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4, 19,
 SET_SER_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 21, 0,
 SET_SER_MUX,      BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 2, 0,
 GET_PEF_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 3, 22,
 SET_PEF_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 22, 0,
//  SET_PEF_ENABLE,  BMC_SA, PUBLIC_BUS, NETFN_APP, BMC_LUN, 4, 0, /*old*/
 GET_DEVSDR_INFO,  BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 0, 6,
 GET_DEVICE_SDR,   BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 6, 18,
 RESERVE_DEVSDR_REP,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 0, 2,
 GET_SENSOR_READING,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 1, 4,
 GET_SENSOR_READING_FACTORS,BMC_SA,PUBLIC_BUS, NETFN_SEVT, BMC_LUN, 2, 7,
 GET_SENSOR_TYPE,  BMC_SA, PUBLIC_BUS, NETFN_SEVT,    BMC_LUN, 1, 2,
 GET_SENSOR_THRESHOLD,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 1, 7,
 SET_SENSOR_THRESHOLD,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 8, 0,
 GET_SENSOR_HYSTERESIS,BMC_SA,PUBLIC_BUS, NETFN_SEVT, BMC_LUN, 2, 2,
 SET_SENSOR_HYSTERESIS,BMC_SA,PUBLIC_BUS, NETFN_SEVT, BMC_LUN, 4, 0,
 GET_SDR,          BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 6, 18, /*full=63*/
 GET_SDR_REPINFO,  BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0, 14,
 RESERVE_SDR_REP,  BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 0, 2,
 GET_FRU_INV_AREA, BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 1, 3,
 READ_FRU_DATA,    BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 4, 18,
 WRITE_FRU_DATA,   BMC_SA, PUBLIC_BUS, NETFN_STOR,  BMC_LUN, 20 /*3+N(17)*/, 1,
 GET_DEVICE_ID,    BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0, 15,
 SET_USER_ACCESS,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 4, 0,
 GET_USER_ACCESS,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 2, 4,
 GET_USER_NAME,    BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 1, 16,
 SET_USER_NAME,    BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 17, 0,
 SET_USER_PASSWORD,BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 18, 0,
 MASTER_WRITE_READ,BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 4 /*or 3*/, 1,
 GET_SYSTEM_GUID,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0, 16,
 WATCHDOG_GET,     BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0, 8,
 WATCHDOG_SET,     BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 6, 0,
 WATCHDOG_RESET,   BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0, 0,
 CHASSIS_CTL,      BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 1, 0,
 CHASSIS_IDENTIFY, BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 1, 0,
 GET_POWERON_HOURS,BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 0, 0,
 SET_BOOT_OPTIONS, BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 19, 0,
 GET_BOOT_OPTIONS, BMC_SA, PUBLIC_BUS, NETFN_CHAS,  BMC_LUN, 3, 18,
 ACTIVATE_SOL,     BMC_SA, PUBLIC_BUS, NETFN_SOL,   BMC_LUN, 0, 0,
 SET_SOL_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SOL,   BMC_LUN, 3, 0,
 GET_SOL_CONFIG,   BMC_SA, PUBLIC_BUS, NETFN_SOL,   BMC_LUN, 4, 2,
 ACTIVATE_SOL2,    BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 0, 0,
 SET_SOL_CONFIG2,  BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 3, 0,
 GET_SOL_CONFIG2,  BMC_SA, PUBLIC_BUS, NETFN_TRANS, BMC_LUN, 4, 2,
 GET_SEVT_ENABLE,  BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 1, 5,
 SET_SEVT_ENABLE,  BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 6, 0,
 REARM_SENSOR,     BMC_SA, PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 6, 0,
 READ_EVENT_MSGBUF,BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 0, 16,
 GET_EVENT_RECEIVER,BMC_SA,PUBLIC_BUS, NETFN_SEVT,  BMC_LUN, 0, 2,
 GET_CHANNEL_INFO, BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 1, 9,
 SET_CHANNEL_ACC,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 3, 0,
 GET_CHANNEL_ACC,  BMC_SA, PUBLIC_BUS, NETFN_APP,   BMC_LUN, 2, 1 };

/* Subroutine definitions for each driver */
extern int ipmi_cmd_ia(ushort cmd, uchar *pdata, uchar sdata, 
	 	uchar *presp, int *sresp, uchar *pcc, char fdebugcmd);
extern int ipmi_open_ia(char fdebug);
extern int ipmi_close_ia(void);
#ifndef GNU_LAN
#include "ipmilan2.h"  /*includes ipmilan.h also*/
#endif
#ifndef WIN32
extern int ipmicmd_mv(uchar cmd, uchar netfn, uchar lun, uchar sa, uchar bus,
		uchar *pdata, uchar sdata, uchar *presp, int sresp, int *rlen);
extern int ipmi_cmd_mv(ushort cmd, uchar *pdata, uchar sdata, uchar *presp, 
			int *sresp, uchar *pcc, char fdebugcmd);
extern int ipmi_open_mv(char fdebug);
extern int ipmi_close_mv(void);
extern int ipmi_open_ld(char fdebug);
extern int ipmi_close_ld(void);
extern int ipmicmd_ld( uchar cmd, uchar netfn, uchar lun, uchar sa, uchar bus,
                uchar *pdata, int sdata, uchar *presp, int *sresp,
                uchar *pcc, char fdebugcmd);
extern int ipmi_cmd_ld(ushort cmd, uchar *pdata, uchar sdata, uchar *presp, 
			int *sresp, uchar *pcc, char fdebugcmd);
extern int ipmi_open_gnu(char *node, char *user, char *pswd, int fdebug);
extern int ipmi_close_gnu(char *node);
extern int ipmicmd_gnu(char *node, uchar cmd, uchar netfn, uchar lun, uchar sa,
		uchar *pdata, uchar sdata, uchar *presp, int *sresp, 
		uchar *pcc, char fdebugcmd);
extern int ipmi_cmd_gnu(char *node, ushort cmd, uchar *pdata, uchar sdata, 
	 	uchar *presp, int *sresp, uchar *pcc, char fdebugcmd);
extern int ipmi_open_direct(char fdebug);
extern int ipmi_close_direct(void);
extern int ipmi_cmd_direct(ushort cmd, uchar *pdata, uchar sdata, uchar *presp, 
			int *sresp, uchar *pcc, char fdebugcmd);
#endif
extern int fd_wait(int fd, int nsec, int usec);

/* Global Data */
int fDriverTyp = 0;   /* 1= IMB driver,   2= VA driver,  3= MV open driver,  */
                      /* 4= GNU FreeIPMI, 5= LanDesk,    6= builtin IPMI LAN */
		      /* 7= direct KCS,   8= direct SMB, 9= IPMI LAN v2.0 */
#define DRV_SMB   8
#define DRV_LAN2  9
int fipmi_lan = 0;    
FILE *fperr = NULL;
FILE *fpdbg = NULL;
char *gnode = "localhost";  /*used for gnu FreeIPMI, util option -N */
char *guser = NULL;
char *gpswd = NULL;
int gshutdown  = 0;
int gauth_type  = IPMI_SESSION_AUTHTYPE_MD5; /*if 0, use any: MD5, MD2, etc.*/
int gpriv_level = IPMI_PRIV_LEVEL_ADMIN;
uchar my_devid[20];  /*saved devid data, only needs 16 bytes*/

typedef struct {
  uchar adrtype;
  uchar sa;
  uchar bus;
  uchar lun;
  uchar capab;
} mc_info;
mc_info  bmc = { ADDR_SMI,  BMC_SA, PUBLIC_BUS, BMC_LUN, 0x8F }; /*BMC via SMI*/
mc_info  mc2 = { ADDR_IPMB, BMC_SA, PUBLIC_BUS, BMC_LUN, 0x4F }; /*IPMB */
static mc_info  *mc  = &bmc;

/* From IPMI v1.5 spec, Table 5-2 Completion Codes */
#define NUMCC    25
struct {
	uchar code;
	char *mesg;
	}  cc_mesg[NUMCC] = { 
0x00, "Command completed successfully",
0x80, "Unsupported Feature",
0xC0, "Node Busy",
0xC1, "Invalid Command",
0xC2, "Command invalid for given LUN",
0xC3, "Timeout while processing command",
0xC4, "Out of space",
0xC5, "Invalid Reservation ID, or cancelled",
0xC6, "Request data truncated",
0xC7, "Request data length invalid",
0xC8, "Request data field length limit exceeded",
0xC9, "Parameter out of range",
0xCA, "Cannot return requested number of data bytes",
0xCB, "Requested sensor, data, or record not present",
0xCC, "Invalid data field in request",
0xCD, "Command illegal for this sensor/record type",
0xCE, "Command response could not be provided",
0xCF, "Cannot execute duplicated request",
0xD0, "SDR Repository in update mode, no response",
0xD1, "Device in firmware update mode, no response",
0xD2, "BMC initialization in progress, no response",
0xD3, "Destination unavailable",
0xD4, "Cannot execute command. Insufficient privilege level",
0xD5, "Cannot execute command. Request parameters not supported",
0xFF, "Unspecified error"
};

char * decode_cc(ushort icmd, uchar cc)
{
   char *pmsg;
   int i;
   for (i = 0; i < NUMCC; i++) {
	if (cc == cc_mesg[i].code) break;
   }
   if (i == NUMCC) i--;  /* if not found, use Unspecified */
   if ((icmd == READ_EVENT_MSGBUF) && (cc == 0x80))
        pmsg = "no data available (queue/buffer empty)";
   else pmsg = cc_mesg[i].mesg;
   return(pmsg);
}

int get_cmd_rslen(uchar cmd, uchar netfn)
{			/* used by ipmicmd_gnu */
   int rslen = 0;
   int i;
   ushort cmdkey;
   cmdkey = cmd | (netfn << 8);
   for (i = 0; i < NCMDS; i++) {
      if (ipmi_cmds[i].cmdtyp == cmdkey) {
	rslen = ipmi_cmds[i].rslen;
	break;
	}
   }
   return(rslen);
} /*end get_cmd_rslen()*/

static struct {
     int idx;
     char *tag;
   } drv_types[NDRIVERS] = {
  { DRV_IMB,  "imb" },
  { DRV_VA,   "va" },
  { DRV_MV,   "open" },
  { DRV_GNU,  "gnu" },
  { DRV_LD,   "landesk" },
  { DRV_LAN2, "lan2" },
  { DRV_LAN,  "lan" },
  { DRV_KCS,  "kcs" },
  { DRV_SMB,  "smb" }
};

char *show_driver_type(int idx)
{
   int i;
   char *tag;
   for (i = 0; i < NDRIVERS; i++)
   {
      if (drv_types[i].idx == idx) {
         tag = drv_types[i].tag;
         break;
      }
   }
   if (i >= NDRIVERS) {  /*not found*/
      tag = "unknown";
   }
   return(tag);
}

int set_driver_type(char *tag)
{
   int rv = 0;
   int i;
   for (i = 0; i < NDRIVERS; i++)
   {
      if (strcmp(drv_types[i].tag, tag) == 0) {
         fDriverTyp = drv_types[i].idx;
         break;
      }
   }
   if (i >= NDRIVERS) {  /*not found*/
      fDriverTyp = 0;  /*not set yet, so detect*/
      rv = 1;
      // if (fdebugcmd) 
      {
         printf("Invalid argument, valid driver types are:\n");
         for (i = 0; i < NDRIVERS; i++)
            printf("\t%s\n",drv_types[i].tag);
      }
   }
   return(rv);
}

int get_driver_type(void)
{
    return(fDriverTyp);
}

int nodeislocal(char *nodename)
{
   if (nodename == NULL) return 1;
   if (nodename[0] == 0) return 1;
   if (strcmp(nodename,"localhost") == 0) return 1;
   return 0;
}

/* 
 * ipmi_open
 * This is called by ipmi_cmd and ipmicmd_raw if a session has not
 * yet been opened (fDriverTyp == DRV_UNKNOWN).
 * The order to try these drivers could be customized for specific
 * environments by modifying this routine.
 */
int ipmi_open(char fdebugcmd)
{
   int rc = 0;
   fperr = stderr;
   fpdbg = stdout;

   if (!nodeislocal(gnode)) fipmi_lan = 1;
   if (fdebugcmd) printf("ipmi_open: driver type = %s\n",
			show_driver_type(fDriverTyp));
   /* first time, so try each */
   if (fipmi_lan) {
#ifdef GNU_LAN
#ifdef WIN32
	    return(-1);
#else
	    rc = ipmi_open_gnu(gnode,guser,gpswd,fdebugcmd); 
	    fDriverTyp = DRV_GNU;
#endif
#else   /*not GNU_LAN*/
      /* Try IPMI LAN 1.5 first */
      rc = ipmi_open_lan(gnode,guser,gpswd,fdebugcmd); 
      fDriverTyp = DRV_LAN;
      if (rc == LAN_ERR_V2) { 
         /* Use IPMI LAN 2.0 if indicated */
         rc = ipmi_open_lan2(gnode,guser,gpswd,fdebugcmd); 
         fDriverTyp = DRV_LAN2;
      }
#endif  /* GNU_LAN*/
   } else {  /* local, not lan */
#ifdef WIN32
	rc = ipmi_open_ia(fdebugcmd);
	if (rc == ACCESS_OK) fDriverTyp = DRV_IMB; 
#else
	if ((rc = ipmi_open_ld(fdebugcmd)) == ACCESS_OK) {
		fDriverTyp = DRV_LD;
		ipmi_close_ld();
	} else if ((rc = ipmi_open_mv(fdebugcmd)) == ACCESS_OK) {
		fDriverTyp = DRV_MV;
		/* ipmi_close_mv(); * leave it open until explicit close */
	} else if ((rc = ipmi_open_ia(fdebugcmd)) == ACCESS_OK) {
		fDriverTyp = DRV_IMB;
	} else if ((rc = ipmi_open_va(fdebugcmd)) == ACCESS_OK) {
		fDriverTyp = DRV_VA;
		ipmi_close_va();
	} else if ((rc = ipmi_open_gnu(gnode,guser,gpswd,fdebugcmd)) == ACCESS_OK) {
		fDriverTyp = DRV_GNU;
	} else if ((rc = ipmi_open_direct(fdebugcmd)) == ACCESS_OK) {
		/* set to either DRV_KCS or DRV_SMB */
	} else rc = ERR_NO_DRV;
#endif

   }  /*endelse local, not lan*/
   if (fdebugcmd) printf("ipmi_open rc = %d type = %s\n",rc,
			show_driver_type(fDriverTyp));
   return (rc);
}

int ipmi_close_(void)
{
   int rc = 0;
   switch (fDriverTyp)
   {
	case DRV_IMB: rc = ipmi_close_ia(); break;
#ifndef WIN32
	case DRV_VA:  rc = ipmi_close_va(); break;
	case DRV_MV:  rc = ipmi_close_mv(); break;
	case DRV_GNU: rc = ipmi_close_gnu(gnode); break;
	case DRV_LD:  rc = ipmi_close_ld(); break;
	case DRV_KCS: rc = ipmi_close_direct(); break;
	case DRV_SMB: rc = ipmi_close_direct(); break;
#endif
#ifndef GNU_LAN
	case DRV_LAN:  rc = ipmi_close_lan(gnode); break;
	case DRV_LAN2: rc = ipmi_close_lan2(gnode); break;
#endif
	default:  break;
   }  /*end switch*/
   return (rc);
}

#ifdef MOVED_TO_IPMILAN_C
/* 
 * ipmi_cmd_ipmb()
 *
 * cksum() is defined in ipmilan.c 
 * Also, iseq needs to be the same as ipmi_hdr.swseq if
 * the session is ipmilan, so this routine was moved.
 * However, ipmi_cmd_ipmb() will also work ok via 
 * ipmi_cmdraw() for local commands.
 */
int ipmi_cmd_ipmb(uchar cmd, uchar netfn, uchar sa, uchar bus, uchar lun,
		uchar *pdata, uchar sdata, uchar *presp,
                int *sresp, uchar *pcc, char fdebugcmd)
{
    int rc, i, j;
    uchar idata[MAX_BUFFER_SIZE];
    int ilen, rlen;
    static uchar iseq = 0;

    ++iseq;
    i = 0;
    idata[i++] = bus;
    j = i;
    idata[i++] = sa;
    idata[i++] = (netfn << 2) | (lun & 0x03);
    idata[i++] = cksum(&idata[j],2);
    j = i;
    idata[i++] = BMC_SA;
    idata[i++] = (iseq << 2) | 0x02;  /*Seq num, SMS Message LUN*/
    idata[i++] = cmd;
    memcpy(&idata[i],pdata,sdata);
    i += sdata;
    idata[i] = cksum(&idata[j],(i - j));
    ilen = ++i;
    rc = ipmi_cmdraw(IPMB_SEND_MESSAGE,NETFN_APP, BMC_SA,PUBLIC_BUS,BMC_LUN,
                  idata, ilen, presp, &rlen, pcc, fdebugcmd);
    if (fdebugcmd) {
       if (rc == 0 && *pcc == 0) 
            dump_buf("ipmb sendmsg ok",presp,rlen,0);
       else fprintf(fpdbg,"ipmb sendmsg error %d, cc %x\n",rc,*pcc);
    }
    if (rc != 0 || *pcc != 0) return(rc);
    /* sent ok, now issue a GETMESSAGE_CMD to get the response. */
    for (i = 0; i < 5; i++)
    {
       rc = ipmi_cmdraw(IPMB_GET_MESSAGE,NETFN_APP, BMC_SA,PUBLIC_BUS,BMC_LUN,
                        idata, 0, presp, sresp, pcc, fdebugcmd);
       if (fdebugcmd) printf("ipmb get_message rc=%d cc=%x\n",rc,*pcc);
       if (rc == 0x80 || *pcc == 0x80) /*retry*/;
       else if (rc == 0x83 || *pcc == 0x83) /*retry*/;
       else break;       /* done w success or error */
       fd_wait(0,0,10);  /* wait 1 msec before retry */
    }
    if (fdebugcmd) {
       if (rc == 0 && *pcc == 0) 
            dump_buf("ipmb getmsg ok",presp,*sresp,0);
       else fprintf(fpdbg,"ipmb getmsg error %d, cc %x\n",rc,*pcc);
    }
    return(rc);
}
#endif

/* 
 * ipmi_cmdraw()
 *
 * This routine can be used to dynamically invoke commands that are not
 * already defined in the ipmi_cmds array by filling in those values 
 * in the empty slot in that array (at index 0).
 * If this app is multi-threaded, this won't be sufficient, since this 
 * empty array slot could be overwritten.  This routine is only 
 * intended for occasional use.
 * It is otherwise similar to ipmi_cmd().
 */
int ipmi_cmdraw(uchar cmd, uchar netfn, uchar sa, uchar bus, uchar lun,
		uchar *pdata, uchar sdata, uchar *presp,
                int *sresp, uchar *pcc, char fdebugcmd)
{
    int rc;
    ushort icmd;

    fperr = stderr;
    fpdbg = stdout;

    if (fDriverTyp == DRV_UNKNOWN) {   /*first time, so find which one */
        rc = ipmi_open(fdebugcmd);
	if (fdebugcmd) 
		fprintf(fpdbg,"Driver type %s, open rc = %d\n",
			show_driver_type(fDriverTyp),rc);
        if (rc == ERR_NO_DRV && !fipmi_lan) fprintf(fpdbg,
			"Cannot open an IPMI driver: /dev/imb, /dev/ipmi0, "
			"/dev/ipmi/0, \n\t/dev/ipmikcs, /dev/ipmi/kcs, "
			"libfreeipmi, or ldipmi.\n");
        else if (rc != 0) fprintf(fpdbg,"ipmi_open error = %d\n", rc);
    }  /*endif first time*/

    /* Set up the temporary ipmi_cmds array structure.  */
    icmd = (cmd & 0x00ff) | (netfn << 8);
    ipmi_cmds[0].cmdtyp = icmd;
    ipmi_cmds[0].sa    = sa;
    ipmi_cmds[0].bus   = bus;
    ipmi_cmds[0].netfn = netfn;
    ipmi_cmds[0].lun   = lun;
    ipmi_cmds[0].len   = sdata;  /*length of request data*/
    *pcc = 0;

    switch (fDriverTyp)
    {
	case DRV_IMB: 
	   rc = ipmi_cmd_ia(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
#ifndef WIN32
	case DRV_VA: 
	   rc = ipmi_cmd_va(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_MV: 
	   rc = ipmi_cmd_mv(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_GNU: 
	   rc = ipmi_cmd_gnu(gnode,icmd,pdata,sdata,presp,sresp,pcc, fdebugcmd);
	   break;
	case DRV_LD: 
	   rc = ipmi_cmd_ld(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_KCS: 
	   rc = ipmi_cmd_direct(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_SMB: 
	   rc = ipmi_cmd_direct(icmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
#endif
#ifndef GNU_LAN
	case DRV_LAN: 
	   rc = ipmi_cmd_lan(gnode,icmd,pdata,sdata,presp,sresp,pcc,fdebugcmd);
	   break;
	case DRV_LAN2: 
	   rc = ipmi_cmd_lan2(gnode,icmd,pdata,sdata,presp,sresp,pcc,fdebugcmd);
	   break;
#endif
	default:    /* no ipmi driver */
	   rc = ERR_NO_DRV;
	   break;
    }  /*end switch*/

    if ((rc >= 0) && (*pcc != 0) && fdebugcmd) {
          fprintf(fpdbg,"raw ccode %x: %s\n",*pcc,decode_cc(icmd,*pcc));
    }
    /* clear the temp cmd */
    ipmi_cmds[0].cmdtyp = 0;    
    ipmi_cmds[0].sa = BMC_SA;

    return(rc);
}

/* 
 * ipmi_cmd_mc()
 * This uses the mc pointer to route commands via either the SMI or
 * IPMB method to the designated mc.
 * See also ipmi_set_mc and ipmi_restore_mc.
 */
int ipmi_cmd_mc(ushort icmd, uchar *pdata, uchar sdata, uchar *presp,
                int *sresp, uchar *pcc, char fdebugcmd)
{
   uchar cmd, netfn;
   int rv;
 
   cmd = icmd & CMDMASK;
   netfn = (icmd & 0xFF00) >> 8;
   if (mc->adrtype == ADDR_IPMB) {
      rv = ipmi_cmd_ipmb(cmd, netfn, mc->sa, mc->bus, mc->lun,
                       pdata, sdata, presp, sresp, pcc, fdebugcmd);
   } else {    /* use ADDR_SMI */
      rv = ipmi_cmdraw(cmd, netfn, mc->sa, mc->bus, mc->lun,
                      pdata, sdata, presp, sresp, pcc, fdebugcmd);
   }
   return(rv);
}

/* 
 * ipmi_cmd()
 *
 * This is the externally exposed subroutine for commands that
 * are defined in the ipmi_cmds array above.
 * It invokes whichever driver-specific routine is needed (va or ia).
 */
int ipmi_cmd(ushort cmd, uchar *pdata, uchar sdata, uchar *presp,
                int *sresp, uchar *pcc, char fdebugcmd)
{
    int rc;

    fperr = stderr;
    fpdbg = stdout;
#ifdef DEBUG
       { /* if (fdebugcmd) */
		fprintf(fpdbg,"stdin %x, stdout %x, stderr %x\n",
			stdin,stdout,stderr);
	}
#endif
    if (fDriverTyp == DRV_UNKNOWN) {   /*first time, so find which one */
        rc = ipmi_open(fdebugcmd);
	if (fdebugcmd) 
		fprintf(fpdbg,"Driver type %s, open rc = %d\n",
			show_driver_type(fDriverTyp),rc);
        if (rc != 0) {
           if (rc == ERR_NO_DRV && !fipmi_lan) fprintf(fpdbg,
			"Cannot open an IPMI driver: /dev/imb, /dev/ipmi0, "
			"/dev/ipmi/0, \n\t/dev/ipmikcs, /dev/ipmi/kcs, "
			"libfreeipmi, or ldipmi.\n");
           else fprintf(fpdbg,"ipmi_open error = %d\n", rc);
           return(rc);
        }
    }  /*endif first time*/

    switch (fDriverTyp)
    {
	case DRV_IMB: 
	   rc = ipmi_cmd_ia(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
#ifndef WIN32
	case DRV_VA: 
	   rc = ipmi_cmd_va(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_MV: 
	   rc = ipmi_cmd_mv(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_GNU: 
	   rc = ipmi_cmd_gnu(gnode,cmd,pdata,sdata,presp,sresp,pcc, fdebugcmd);
	   break;
	case DRV_LD: 
	   rc = ipmi_cmd_ld(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_KCS: 
	   rc = ipmi_cmd_direct(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
	case DRV_SMB: 
	   rc = ipmi_cmd_direct(cmd, pdata,sdata, presp,sresp, pcc, fdebugcmd);
	   break;
#endif
#ifndef GNU_LAN
	case DRV_LAN: 
	   rc = ipmi_cmd_lan(gnode,cmd,pdata,sdata,presp,sresp,pcc,fdebugcmd);
	   break;
	case DRV_LAN2: 
	   rc = ipmi_cmd_lan2(gnode,cmd,pdata,sdata,presp,sresp,pcc,fdebugcmd);
	   break;
#endif
	default:    /* no ipmi driver */
	   rc = ERR_NO_DRV;
	   break;
    }  /*end switch*/
    if ((rc >= 0) && (*pcc != 0) && fdebugcmd) 
          fprintf(fpdbg,"ccode %x: %s\n",*pcc,decode_cc(cmd,*pcc));

    return(rc);
}

int ipmi_getpicmg(uchar *presp, int sresp, char fdebug)
{
   uchar idata[2];
   int rc; uchar cc;

   /* check that sresp is big enough */
   if (sresp < 4) return(-3);
   idata[0] = PICMG_ID;
   rc = ipmi_cmdraw(PICMG_GET_PROPERTIES, NETFN_PICMG,
		    BMC_SA, PUBLIC_BUS, BMC_LUN, 
		    idata, 1, presp,&sresp, &cc, fdebug);
   if (rc != ACCESS_OK) return(rc);
   if (cc != 0) return(cc);
   return(ACCESS_OK);   /* success */
}

int ipmi_getdeviceid(uchar *presp, int sresp, char fdebug)
{
   int rc; uchar cc;
   /* check that sresp is big enough (default is 15 bytes for Langley)*/
   if (sresp < 15) return(-3);
   rc = ipmi_cmd(GET_DEVICE_ID, NULL, 0, presp,&sresp, &cc, fdebug);
   if (rc != ACCESS_OK) return(rc);
   if (cc != 0) return(cc);
   if (sresp > sizeof(my_devid)) sresp = sizeof(my_devid);
   memcpy(my_devid,presp,sresp); /* save device id for later use */
   return(ACCESS_OK);   /* success */
}

void ipmi_set_mc(uchar bus, uchar sa, uchar lun, uchar type) 
{
   mc = &mc2;
   mc->bus  = bus;
   mc->sa   = sa;
   mc->lun  = lun;
   mc->adrtype = type; /* ADDR_SMI or ADDR_IPMB */
   return;
}

void ipmi_restore_mc(void)
{
   mc = &bmc;
   return;
}

/* Need our own copy of strdup(), named strdup_(), since Windows does
 * not have the same subroutine. */
char * strdup_(const char *instr)
{
    char *newstr;
    newstr = malloc(strlen(instr)+1);
    if (newstr != NULL) strcpy(newstr,instr);
    return (newstr);
}

/*
 * htoi
 * Almost all of the utilities use this subroutine 
 * Input:  a 2 character string of hex digits.
 * Output: a hex byte.
 */
uchar htoi(uchar *inhex)
{
   // char rghex[16] = "0123456789ABCDEF";
   uchar val;
   uchar c;
   if (inhex[1] == 0) { /* short string, one char */
      c = inhex[0] & 0x5f;  /* force cap */
      if (c > '9') c += 9;  /* c >= 'A' */
      val = (c & 0x0f);
   } else {
      c = inhex[0] & 0x5f;  /* force cap */
      if (c > '9') c += 9;  /* c >= 'A' */
      val = (c & 0x0f) << 4;
      c = inhex[1] & 0x5f;  /* force cap */
      if (c > '9') c += 9;  /* c >= 'A' */
      val += (c & 0x0f);
   }
   return(val);
}

void dump_buf(char *tag, uchar *pbuf, int sz, char fshowascii)
{
   uchar line[17];
   uchar a;
   int i, j;

   line[0] = 0; line[16] = 0;
   j = 0;
   fprintf(fpdbg,"%s (len=%d): ", tag,sz);
   for (i = 0; i < sz; i++) {
      if (i % 16 == 0) { j = 0; fprintf(fpdbg,"%s\n  %04d: ",line,i); }
      if (fshowascii) {
         a = pbuf[i];
         if (a < 0x20 || a > 0x7f) a = '.';
         line[j++] = a;
      }
      fprintf(fpdbg,"%02x ",pbuf[i]);
   }
   fprintf(fpdbg,"%s\n",line);
   return;
} 

#ifdef WIN32
static HANDLE con_in  = INVALID_HANDLE_VALUE; 
static DWORD cmodein;
static DWORD cmodeold;
void tty_setraw(int mode)
{
    // system("@echo off");
    con_in  = GetStdHandle(STD_INPUT_HANDLE);
    GetConsoleMode(con_in,  &cmodein);
    cmodeold = cmodein;
    if (mode == 2) {
       cmodein &= ~(ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT | 
    		   ENABLE_ECHO_INPUT);
    } else {  /* (mode==1) just suppress ECHO */
       cmodein &= ~ENABLE_ECHO_INPUT;
    }
    SetConsoleMode(con_in,  cmodein);
}
void tty_setnormal(int mode)
{
    // system("echo on");
    if (mode == 1) 
       cmodein  |= ENABLE_ECHO_INPUT;
    else 
       cmodein = cmodeold;
    SetConsoleMode(con_in,  cmodein);
}
#else
          /*Linux*/
static struct termios mytty;
static ulong  tty_oldflags;
void tty_setraw(int mode)
{
    int i;
    // system("stty -echo");
    i = tcgetattr(STDIN_FILENO, &mytty);
    if (i == 0) {
        tty_oldflags = mytty.c_lflag;
        if (mode == 2)   /*raw mode*/
           mytty.c_lflag &= ~(ICANON | ISIG | ECHO);
        else   /* (mode==1) just suppress ECHO */
           mytty.c_lflag &= ~ECHO;
        tcsetattr(STDIN_FILENO, TCSANOW, &mytty);
    }
}
void tty_setnormal(int mode)
{
    // system("stty echo");
    if (mode == 1) 
       mytty.c_lflag |= ECHO;
    else  /*(mode==2)*/
       mytty.c_lflag = tty_oldflags;
    tcsetattr(STDIN_FILENO, TCSANOW, &mytty);
}
#endif

static char *getline(char *prompt, char fwipe)
{
   /* getline is the same format as readline, but much simpler, and portable. */
   static char linebuf[128];
   int c, i;

   if (prompt != NULL) printf("%s\n",prompt);
   if (fwipe) tty_setraw(1); 
   for (i = 0; i < (sizeof(linebuf)-1); i++)
   {
       c = getc(stdin);
       if (c == EOF)  break;
       if (c == '\n') break;
       if ((c < 0x20) || (c > 0x7F)) break;  /*out of bounds for ASCII */
       linebuf[i] = c & 0xff;
   }
   linebuf[i] = 0;
   if (fwipe) {
       for (c = 0; c < i; c++) putc('*',stdout);
       putc('\n',stdout);
       tty_setnormal(1);
   }
   if (i == 0) return NULL;
   return(linebuf);
}

static char *getline_wipe(char *prompt)
{
   return(getline(prompt,1));
}

void parse_lan_options(char c, char *optarg, char fdebugcmd)
{
   int i;

   switch(c) 
   {
          case 'F':      /* force driver type */
                i = set_driver_type(optarg);
                break;
          case 'T':      /* auth type */
                i = atoi(optarg);
                if (i > 0 && i <= 5) gauth_type = i;
                break;
          case 'V':      /* priv level */
                i = atoi(optarg);
                if (i > 0 && i <= 5) gpriv_level = i;
                break;
          case 'N': gnode  = optarg;   /* nodename */
                    fipmi_lan = 1;
                    break;
          case 'U':
                guser = strdup_(optarg);  /*remote username */
                if (!guser)  perror("strdup_");
                else {  /* Hide password from 'ps' */
                        memset(optarg, ' ', strlen(optarg));
                }
                break;
          case 'R':
          case 'P':
                gpswd = strdup_(optarg);  /*remote password */
                if (!gpswd) perror("strdup_");
                else {  /* Hide password from 'ps' */
                        memset(optarg, ' ', strlen(optarg));
                }
                break;
          case 'E':      /* get password from IPMI_PASSWORD environment var */
                gpswd = getenv("IPMI_PASSWORD");
                if (gpswd == NULL) perror("getenv(IPMI_PASSWORD)");
                else {
                   if (strlen(gpswd) > 16) gpswd[16] = 0;
                   if (fdebugcmd) printf("using IPMI_PASSWORD\n");
                }
                break;
          case 'Y':      /* prompt for remote password */
		gpswd = getline_wipe("Enter IPMI LAN Password: ");
                break;
          default:
                break;
   }
   return;
}  /*end parse_lan_options*/

void print_lan_opt_usage(void)
{
    printf("       -N node  Nodename of target system\n");
    printf("       -U user  Username for remote node\n");
    printf("       -P/-R pswd  Remote Password\n");
    printf("       -E   use password from Environment IPMI_PASSWORD\n");
    printf("       -T 1 use auth Type: 1=MD2, 2=MD5(default)\n");
    printf("       -V 2 use priVilege level: 2=user, 4=admin(default)\n");
    printf("       -F   force driver type (e.g. imb, lan2)\n");
    printf("       -Y   prompt for remote password\n");
}  /*end parse_lan_options*/

/* end ipmicmd.c */
