/*
 * isolconsole.c
 *      IPMI Serial-Over-LAN console application
 *
 * Author:  Andy Cress  arcress@users.sourceforge.net
 * Copyright (c) 2006 Intel Corporation. 
 *
 * 10/12/06 Andy Cress - created, just a framework for now
 * 02/08/07 Andy Cress - now working for IPMI LAN 2.0 SOL in Linux,
 *                       still stdio problems in Windows
 * 02/23/07 Andy Cress - fixed Windows stdin, still Windows output issues,
 *                       need more in xlate_vt100_ansi().
 *
 * NOTE: There are a few other options for Serial-Over-LAN console 
 * applications:
 *   ipmitool.sf.net (has v2 sol console capability for Linux)
 *   Intel dpccli (CLI console for Linux and Windows)
 *   Intel ISM console (for Windows)
 *   Intel System Management Software (LANDesk) console 
 * The Intel applications support both IPMI 1.5 and 2.0 SOL 
 * protocols.
 */
/*M*
Copyright (c) 2006, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include <winsock.h>
#include <io.h>
#include <conio.h>
#include "getopt.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#endif
#include <string.h>
#include <time.h>
#include "ipmicmd.h"
#include "ipmilanplus.h"
 
#define ACTIVATE_PAYLOAD    0x48
#define DEACTIVATE_PAYLOAD  0x49
#define PAYLOAD_TYPE_SOL    0x01
#define SOL_SERIAL_ALERT_MASK_DEFERRED 0x04
#define SOL_BMC_ASSERTS_CTS_MASK_TRUE  0x00  
#define SOL_BMC_ASSERTS_CTS_MASK_FALSE 0x02
#define AUTHTYPE_RMCP_PLUS  0x06  
#define RV_END   -2

typedef struct {
  int type;
  int len;
  char *data;
  } SOL_RSP_PKT;

extern int  lan_get_fd(void);           /*from ipmilan.c*/
extern int  lan_send_sol( uchar *payload, int len, SOL_RSP_PKT *rsp);
extern int  lan_recv_sol( SOL_RSP_PKT *rsp );
extern int  lan_keepalive(int type);
extern int  lan2_get_fd(void);          /*from ipmilanplus.c*/
extern int  lan2_send_sol( uchar *payload, int len, SOL_RSP_PKT *rsp);
extern int  lan2_recv_sol( SOL_RSP_PKT *rsp );
extern void lan2_set_sol_data(int insize, int outsize, int port, void*hnd);
extern int  lan2_keepalive(int type);
extern void tty_setraw(int mode);      /*from ipmicmd.c*/
extern void tty_setnormal(int mode);   /*from ipmicmd.c*/

/*
 * Global variables 
 */
static char * progver   = "1.0";
static char * progname  = "isolconsole";
static char  fdebug        = 0;
static char  fpicmg        = 0;
static char  factivate     = 0;
static char  fdeactivate   = 0;
static char  bDriver = 0;
static uchar bSolVer       = 0;
static uchar bKeepAlive    = 2;
static uchar ipmi_maj = 0;
static uchar ipmi_min = 0;
static int   mfg      = 0;  /* Manufacturer IANA vendor id */
static uchar fSolEncryption     = 1;
static uchar fSolAuthentication = 1;
static char  esc_ch        = '~';  /* escape char ('~') */
static int done = 0;

#ifdef WIN32
/*======== start Windows =============================*/
    /* Windows os_read, os_select and supporting code */
#define OUTBUF_SZ   128
#define INBUF_SZ    64
static uchar rg_stdin[INBUF_SZ];
static int   n_stdin = 0;
/* extern int _cdecl _getche(void);   * see conio.h */

#define  VT_ESC   0x1B
static struct { 
   int  vt_sz;
   char vt[6];
   char an[6];
} esc_vt_an[9] = {
  5, {VT_ESC,"[PA"},  { "" }, /*cursor up    */
  5, {VT_ESC,"[PB"},  { "" }, /*cursor down  */
  5, {VT_ESC,"[PC"},  { "" }, /*cursor right */
  5, {VT_ESC,"[PD"},  { "" }, /*cursor left  */
  5, {VT_ESC,"[Pm"},  { "" }, /*char attributes*/
  2, {VT_ESC,"E"},   { "" },  /*next line */
  3, {VT_ESC,"("},   { "" },  /*character set*/
  4, {VT_ESC,"[K"},  { "" },  /*erase line*/
  4, {VT_ESC,"[J"},  { "" }   /*erase screen*/
};

static int IsVtEsc(uchar c, uchar *buf, int idx)
{
   int ret = 0;
   switch (idx) {
       case 0:  if (c == VT_ESC) ret = 1;  break;
       case 1:  
          if (c == '[') ret = 1;  
          else if (c == '(') ret = 1;
          else if (c == 'E') ret = 1;
          break;
       case 2:
          if (c == 'P') ret = 1;
          else if (c == 'K') ret = 1;
          else if (c == 'J') ret = 1;
          break;
       case 5:
          if (buf[1] != '[' || buf[2] != 'P') 
              return(0); /* only ESC [P has >5 chars */
       case 4:
       case 3:
          if (c >= '0' && c <= '9') ret = 1;
          else if (c >= 'A' && c <= 'D') ret = 1;
          else if (c == 'm') ret = 1;
          else if (c == 'K') ret = 1;
          else if (c == 'J') ret = 1;
          else if (c == ';') ret = 1;
          break;
       default:
          break;
   }
   return(ret);
}

static void xlate_vt100_ansi(uchar *pin, int ilen, uchar *pout, int *olen)
{
    int i, j;
    static int iesc = 0;
    static uchar escbuf[16];

    /* Translate vt100 escape sequences to ansi here. */
    j = 0; 
    for (i = 0; i < ilen; i++) {
       if (j >= OUTBUF_SZ) break;
       if (IsVtEsc(pin[i],escbuf,iesc)) {  /*vt100 esc sequence*/
          escbuf[iesc++] = pin[i];
          /* TODO: check if valid/complete sequence, and translate */
       } else {  /*standard chars pass through*/
          iesc = 0;
          pout[j++] = pin[i];
       }
    }
    *olen = j;
}

DWORD WINAPI input_thread( LPVOID lpParam) 
{
   // HANDLE hstdin;
   DWORD rv = 0;
   int c;

   // hstdin = GetStdHandle(STD_INPUT_HANDLE); 
   while (!done)
   {
      c = _getche();   /*was getc(stdin) */
      if (c != EOF) {
         if (n_stdin < sizeof(rg_stdin)) {
            rg_stdin[n_stdin] = c;
            n_stdin++;
            // if (fdebug) printf("getche() got 0x%02x, %d\n",c,n_stdin);
         } 
         // else if (fdebug) printf("getche() n_stdin=%d > max\n",n_stdin);
      } 
      // else if (fdebug) printf("getche() == EOF\n");
   }
   return(rv);
}

static int os_read(int fd, uchar *buf, int sz)
{
    int len = 0;
    int i;
    if ((fd == 0) && (n_stdin > 0)) {
       for (i = 0; i < n_stdin; i++)
          buf[i] = rg_stdin[i];
       len = n_stdin;
       n_stdin = 0;
    } else {
       // ReadFile(fd, buf, sz, &len, NULL);
       // len = _read(fd, buf, sz);
       len = recv(fd, buf, sz, 0);
    }
    return(len);
}

int os_select(int infd, int sfd, fd_set *read_fds)
{
   int rv = 0;
   struct timeval tv;
   fd_set err_fds;
   int rv0 = -1;
   // int c;

   /* check the socket for new data via select */
   /* Windows select only works on WSA sockets */
   {
      FD_ZERO(read_fds);
      FD_SET(sfd, read_fds);
      FD_ZERO(&err_fds);
      tv.tv_sec =  0;
      tv.tv_usec = 50000; /* 50 ms */
      rv = select(sfd + 1, read_fds, NULL, NULL, &tv);
      if (rv < 0) {
         rv = -(WSAGetLastError());
         if (fdebug) printf("select(%d) error %d\n",sfd,rv);
      }
      else if (FD_ISSET(sfd, &err_fds)) {
         if (fdebug) printf("select(%d) err_fds rv=%d err=%d\n",sfd, rv, 
			 WSAGetLastError());
      }
      if (rv > 0 && fdebug) printf("select(%d) got socket data %d\n",sfd,rv);
   }  
   if (infd == 0) {  /* check stdin */
      rv0 = -1;
      if (n_stdin > 0) rv0 = 0;
   }
   if (rv0 == 0) {
      FD_SET(infd, read_fds);
      if (rv < 0) rv = 1;
      else rv++;
   }
   return rv;
}
/*======== end Windows ===============================*/
#else
/*======== start Linux ===============================*/
static int os_read(int fd, uchar *buf, int sz)
{
    int len = 0;
    len = read(fd, buf, sz);
    return(len);
}

int os_select(int infd, int sfd, fd_set *read_fds)
{
   int rv = 0;
   struct timeval tv;

   /* Linux handles both stdin & socket via select() */
   FD_ZERO(read_fds);
   FD_SET(infd, read_fds);
   FD_SET(sfd, read_fds);
   tv.tv_sec =  0;
   tv.tv_usec = 500000;
   rv = select(sfd + 1, read_fds, NULL, NULL, &tv);
   return rv;
}
/*======== end Linux =================================*/
#endif

static int sol_send(uchar *buf, int len, SOL_RSP_PKT *rsp)
{
    int rv = 0;
    if (bSolVer == 1) {
       rv = lan_send_sol(buf, len, rsp);
    } else 
       rv = lan2_send_sol(buf, len, rsp);
    return(rv);
}

static int sol_recv( SOL_RSP_PKT *rsp)
{
    int rv = 0;
    if (bSolVer == 1) {
       rv = lan_recv_sol( rsp );
    } else 
       rv = lan2_recv_sol( rsp );
    return(rv);
}

static int  sol_get_fd(void)
{
    int fd = 0;
    if (bSolVer == 1) {
       fd = lan_get_fd();
    } else 
       fd = lan2_get_fd();
    return(fd);
}

#define SOL_TIMEOUT   30
static int sol_keepalive(int type)
{
   int rv = 0;
   static time_t keepalive_start; 
   time_t end; 

   time(&end); 
   if (end - keepalive_start > SOL_TIMEOUT) {
       if (bSolVer == 1) {
           rv = lan_keepalive(bKeepAlive);   
       } else 
           rv = lan2_keepalive(bKeepAlive);   
       time(&keepalive_start);
       if (fdebug) printf("sol_keepalive sent, rv = %d\n",rv);
   }
   return(rv);
}

void sol_output_handler(SOL_RSP_PKT *rsp)
{
    int i;
    uchar c;
    uchar *pdata;
    int len;
    if (rsp &&
        (rsp->type == PAYLOAD_TYPE_SOL))
    {
#ifdef WIN32
         uchar buf[OUTBUF_SZ];
         pdata = buf;
         xlate_vt100_ansi(rsp->data,rsp->len,pdata,&len);
#else
         pdata = rsp->data;
         len   = rsp->len;
#endif
         if (fdebug) printf("sol_output: len = %d\n",len);
         for (i = 0; i < len; i++) {
	      c = pdata[i];
              /* if (c == '\r') continue;  *skip CR '\r' chars*/
              putc(c, stdout);
         }
         fflush(stdout);
    }
}

static int sol_retries = 3;

int sol_input_handler( uchar *input, int ilen)
{
   int rv = 0;
   uchar payload[0x64]; 
   SOL_RSP_PKT  rs;
   int i, length, t;
   uchar c, lastc;

   memset(&payload, 0, sizeof(payload)); 
   length = 0;
   rs.len = 0;
   for (i = 0; i < ilen; ++i)
   {
        c = input[i];

        /* look for escape sequences */
        // if (lastc == '\n' && c == esc_ch) 
        if (c == esc_ch) {  /*'~'*/
           return(RV_END);
        }
#ifdef WIN32
        if (c == 0x0d) /*0x0D = '\r' from Windows Enter*/
            c = 0x0a;  /*0x0A = '\n' newline */
#endif

        /* copy other data to payload */
        payload[length++] = c;
        lastc = c;
   }
   if (length > 0) {
      for (t = 0; t < sol_retries; t++) {
         rv = sol_send(payload, length, &rs);
         if (fdebug) 
		printf("sol_send(%d,%d) rv = %d (0x%02x)\n", t,length,rv,rv);
         if (rv == 0) break;
         os_usleep(0,5000);
      }
      if (rv >= 0 && rs.len != 0) { /*handle the rsp data*/
          if (fdebug) printf("output: handler(%d)\n",rs.len);
          sol_output_handler(&rs);
      }
   }
   return(rv);
}

static int send_activate_15sol(void)
{
   int rv = 0;
   uchar actcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;

   actcmd  = (uchar)(ACTIVATE_SOL & 0x00ff); 
   netfn   = NETFN_SOL; 
   ibuf[0] = 0x01;  
   ibuf[1] = 0x00;  
   ibuf[2] = 0x00;  
   ibuf[3] = 0x00;  
   ibuf[4] = 0x00;  
   // ibuf[5] = 0x00;  
   ilen = 5;
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(actcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (rv == 0 && cc != 0) rv = cc;
   /* switch(cc) { case 0x80: SOL already active on another session
    *              case 0x81: SOL disabled
    *              case 0x82: SOL activation limit reached  }
    */
   if (rv == 0) {  /*success*/
      if (fdebug) dump_buf("sol 15 act_resp",rbuf,rlen,0);
   }
   return(rv);
}

static int send_deactivate_15sol(void)
{
   int rv = -1;

   uchar actcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;

   /* SOL for IPMI 1.5 does not have an explicit deactivate, so
    * use an activate with data0 = 00. 
    */
   actcmd  = (uchar)(ACTIVATE_SOL & 0x00ff); 
   netfn   = NETFN_SOL; 
   ibuf[0] = 0x00;    /*deactivate*/
   ibuf[1] = 0x00;  
   ibuf[2] = 0x00;  
   ibuf[3] = 0x00;  
   ibuf[4] = 0x00;  
   ilen = 5;
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(actcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (rv == 0 && cc != 0) rv = cc;
   if (fdebug) dump_buf("sol 15 deact_resp",rbuf,rlen,0); /* 80 00 32 ff */
   return(rv);
}

int send_activate_sol(uchar verSOL)
{
   int rv = 0;
   int in_payload_size, out_payload_size, udp_port;
   uchar actcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;
   /* activate SOL = 0x01, get SOL status = 0x05 */
   if (verSOL == 1) {
      return(send_activate_15sol());
   } else {   /* use IPMI 2.0 method */
      actcmd  = ACTIVATE_PAYLOAD;
      netfn   = NETFN_APP;  
      ibuf[0] = PAYLOAD_TYPE_SOL;
      ibuf[1] = 1;        /* payload instance */
      ibuf[2] = SOL_SERIAL_ALERT_MASK_DEFERRED;
      if (fSolEncryption)     ibuf[2] |= 0x80;
      if (fSolAuthentication) ibuf[2] |= 0x40;
      if (mfg == VENDOR_INTEL) 
           ibuf[2] |= SOL_BMC_ASSERTS_CTS_MASK_TRUE;
      else ibuf[2] |= SOL_BMC_ASSERTS_CTS_MASK_FALSE;
      ibuf[3] = 0;
      ibuf[4] = 0;
      ibuf[5] = 0;
      ilen = 6;
   }
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(actcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (fdebug) {
      dump_buf("sol act_cmd",ibuf,ilen,0);
      printf("send_activate v2(%x,%x) rv = %d cc = %x\n",actcmd,netfn,rv,cc);
   }
   if (rv >= 0) {
      rv = cc;
      switch(cc) {
         case 0x00: 
	       if (rlen != 12) {
	         printf("Unexpected SOL response data received, len=%d\n",rlen);
	         rv = -1;
               }
               break;
         case 0x80: 
	       printf("SOL payload already active on another session\n");
               break;
         case 0x81: 
	       printf("SOL payload disabled\n");
               break;
         case 0x82: 
	       printf("SOL payload activation limit reached\n");
               break;
         case 0x83: 
	       printf("Cannot activate SOL payload with encryption\n");
               break;
         case 0x84: 
	       printf("Cannot activate SOL payload without encryption\n");
               break;
         default:
	       printf("Cannot activate SOL, ccode = 0x%02x\n",cc);
               break;
      }
   }
   if (rv == 0) {   /* success, use the response data */
      /* only here if response data is from IPMI 2.0 method */
      if (fdebug) dump_buf("sol act_resp",rbuf,rlen,0);
      in_payload_size  = rbuf[4] + (rbuf[5] << 8);
      out_payload_size = rbuf[6] + (rbuf[7] << 8);
      udp_port         = rbuf[8] + (rbuf[9] << 8);
      if (fdebug) printf("activate ok, in=%d out=%d port=%d\n",
			in_payload_size, out_payload_size, udp_port);
      if (bSolVer == 2) 
         lan2_set_sol_data(in_payload_size, out_payload_size, udp_port,
			   (void *)sol_output_handler);
   }
   return(rv);
}

int send_deactivate_sol(uchar verSOL)
{
   int rv = 0;
   uchar deactcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;

   if (verSOL == 1) {
      return(send_deactivate_15sol());
   } else {   /* use IPMI 2.0 method */
      deactcmd  = DEACTIVATE_PAYLOAD;
      netfn   = NETFN_APP;
      ibuf[0] = PAYLOAD_TYPE_SOL;   /* payload type */
      ibuf[1] = 1;                  /* payload instance */
      ibuf[2] = 0;
      ibuf[3] = 0;
      ibuf[4] = 0;
      ibuf[5] = 0;
      ilen = 6;
   }
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(deactcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (fdebug) printf("sol deactivate rv = %d, cc = %x\n",rv,cc);
   if (rv == 0 && cc != 0) rv = cc;
   return(rv);
}


int sol_data_loop( void )
{
   int istdin;
   int rv = 0;
   uchar ibuf[0x64];
   int szibuf;
   int fd;
   int len;
   fd_set read_fds;
#ifdef WIN32
   HANDLE thnd;
#endif

   /* 
    * OK, now loop and listen to 
    *   stdin for user input 
    *   interface fd for incoming SOL data 
    */
   szibuf = sizeof(ibuf);  /* or via intf sol data */
   istdin = fileno(stdin);
   fd = sol_get_fd();
   if (fdebug) printf("stdin = %d, intf->fd = %d\n",istdin,fd);
   printf("[SOL session is running, use '%c' to end session.]\n",esc_ch);

   tty_setraw(2);
#ifdef WIN32
   thnd = CreateThread(NULL, 0, &input_thread, NULL, 0, NULL);
   if (thnd == NULL) { 
       printf("CreateThread error, aborting\n");
       rv = -1; 
       done = 1; 
   }
#endif
   while (done == 0) 
   {
       if (bKeepAlive > 0)   /* send KeepAlive */
           rv = sol_keepalive(bKeepAlive);   

       rv = os_select(istdin,fd, &read_fds);
       if (rv < 0) { 
	   if (fdebug) printf("os_select(%d) error %d\n",fd,rv); 
	   perror("select"); 
           done = 1; 
       } else if (rv > 0) {
           if (fdebug) printf("select rv = %d\n",rv);

           if (FD_ISSET(istdin, &read_fds)) {
               /* input from stdin (user) */
               memset(ibuf,0,szibuf);
               len = os_read(istdin, ibuf, szibuf);
               if (fdebug) printf("input: read len=%d, %02x %02x\n",
					len,ibuf[0], ibuf[1]);
               if (len <= 0) {
                  printf("Error %d reading stdin\n",len);
                  done = 1;
               } else {
                  rv = sol_input_handler(ibuf,len);
                  if (fdebug) printf("input: handler rv = %d\n",rv);
                  if (rv < 0) {
                      done = 1;
                      if (rv == RV_END) 
			   printf("%s exit via user input\n",progname);
                      else printf("Error %d processing input\n",rv);
                  }
               }
           }  /*endif stdin*/

           if (FD_ISSET(fd, &read_fds)) {
               SOL_RSP_PKT rs;
               rs.len = 0;
               /* output from BMC SOL */
               rv = sol_recv(&rs);
#ifdef WIN32
		/* Windows sometimes gets rv = -1 here */
               if (rv == -1) rv = -(WSAGetLastError());
#endif
               if (fdebug) printf("output: recv rv = %d\n",rv);
               if (rv < 0) {
                   printf("Error %d reading SOL socket\n",rv);
                   done = 1;
               } else sol_output_handler(&rs);
           }  /*endif fd*/
       }  /*endif select rv > 0 */
   }  /*end while*/
#ifdef WIN32
   CloseHandle(thnd);
#endif
   tty_setnormal(2);
   return(rv);
}

static void show_usage(void)
{
                printf("Usage: %s [-adx -NUPREFTVY]\n", progname);
                printf(" where -a      activate SOL console session\n");
                printf("       -d      deactivate SOL console session\n");
                printf("       -e      Encryption off for SOL session\n");
                printf("       -x      show eXtra debug messages\n");
		print_lan_opt_usage();
}

#ifdef METACOMMAND
int i_sol(int argc, char **argv)
#else
#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
#endif
{
   int ret = 0;
   int i;
   uchar devrec[16];
   uchar done = 0;
   char c;

   printf("%s ver %s\n", progname,progver);

   while ( (c = getopt( argc, argv,"adexEF:T:V:N:P:R:U:Y?")) != EOF ) 
      switch(c) {
          case 'a': factivate = 1;   break;    /*activate*/
          case 'd': fdeactivate = 1; break;    /*deactivate*/
          case 'e': fSolEncryption = 0; break; /*encryption off*/
          case 'x': fdebug = 1;      break;    /*debug messages */
          case 'N':    /* nodename */
          case 'U':    /* remote username */
          case 'P':    /* remote password */
          case 'R':    /* remote password */
          case 'E':    /* get password from IPMI_PASSWORD environment var */
          case 'F':    /* force driver type */
          case 'T':    /* auth type */
          case 'V':    /* priv level */
          case 'Y':    /* prompt for remote password */
                parse_lan_options(c,optarg,fdebug);
                break;
	  default:
                show_usage();
                exit(1);
      }

   if (fipmi_lan == 0) {  /* no node specified */
      printf("Serial-Over-Lan Console requires a -N nodename\n");
      exit(1);
   }

   if (factivate == 0 && fdeactivate == 0) {
      show_usage();
      printf("Error: Need either -a or -d for sol\n");
      exit(1);
   }

   i = get_driver_type(); /*see if user explictly set type*/
   if (i == 0) bDriver = 0;  /*no driver type specified*/
   else bDriver = i;

   ret = ipmi_getdeviceid(devrec,16,fdebug);
   if (ret != 0) {
	printf("Cannot do ipmi_getdeviceid, ret = %d\n",ret);
	ipmi_close_();
	exit(1);
   } else {
      ipmi_maj = devrec[4] & 0x0f;
      ipmi_min = devrec[4] >> 4;
      mfg  = devrec[6] + (devrec[7] << 8) + (devrec[8] << 16); 
      printf("-- BMC version %x.%x, IPMI version %d.%d \n",
             devrec[2],  devrec[3], ipmi_maj, ipmi_min);
   }

   ret = ipmi_getpicmg(devrec,16,fdebug);
   if (ret == 0) fpicmg = 1;

   if (fdebug) printf("driver=%d mfg=%x\n",bDriver,mfg); /*++++*/
   /* check for SOL support */
   if (ipmi_maj >= 2) { 
       if ((bDriver == DRV_LAN) && (mfg == VENDOR_INTEL)) {
          /* user specified to use IPMI LAN 1.5 SOL on Intel */
          bSolVer = 1;      /*IPMI 1.5 SOL*/
       } else {
          bSolVer = 2;      /*IPMI 2.0 SOL*/
          if (get_driver_type() == DRV_LAN) { /*now using IPMI LAN 1.5*/
            ipmi_close_();           /*close current IPMI LAN 1.5*/
            i = set_driver_type("lan2"); /*switch to IPMI LAN 2.0*/
          }
       } 
   } else if (ipmi_maj == 1) {
       if (ipmi_min >= 5) bSolVer = 1;  /* IPMI 1.5 */
       else bSolVer = 0;  /* IPMI 1.0 */
   } else bSolVer = 0;
   if (bSolVer == 0) {
       printf("Serial Over Lan not supported for this IPMI version\n");
       exit(1);
   }
   /* May also want to verify that SOL is implemented here */

   /*
    * Spawn a console raw terminal thread now, which will wait for the 
    * "Activating cmd (0x02)" on success 
    * Using globals: gnode,guser,gpswd, gauth_type, gpriv_level
    */
   if (fdeactivate) {
      ret = send_deactivate_sol(bSolVer); 
      if (fdebug) printf("send_deactivate_sol rv = %d\n",ret);
   } else if (factivate) {
      ret = send_activate_sol(bSolVer);
      if (fdebug) printf("send_activate_sol rv = %d\n",ret);
      if (ret == 0) {
         ret = sol_data_loop();
      } /*endif activate ok*/
   }  /*endif do activate */

   ipmi_close_();
   exit (ret);
}  /* end main()*/

/* end isolconsole.c */
