/*
 * showsel.c
 *
 * This tool reads the firmware System Event Log records via IPMI commands.
 * It can use either the Intel /dev/imb driver or VALinux /dev/ipmikcs.
 *
 * Author:  Andy Cress  arcress@users.sourceforge.net
 *
 * Copyright (c) 2001-2005 Intel Corporation. 
 *
 * 10/16/01 Andy Cress - created
 * 10/19/01 Andy Cress - added text for Sensor Types from Table 36-3
 * 10/24/01 Andy Cress - added panic string display for OS Crit Stop
 * 11/01/01 Andy Cress - added logic to write SEL records to syslog
 * 11/15/01 Andy Cress - added logic for more description strings
 * 01/31/02 Andy Cress - isolated dependencies on /dev/imb to ipmi_cmd_ia(),
 *                       added ipmi_cmd_va() for va ipmi driver also,
 *                       added posterrs array for more descriptions.
 * 02/06/02 Andy Cress - fixed bug 279 (extra debug msg at EOF).
 * 03/25/02 Andy Cress - show free space, add -c to clear SEL.
 * 04/11/02 Andy Cress - decode timestamp into readable form
 * 06/14/02 Andy Cress - also show status error in ClearSEL
 * 07/02/02 Andy Cress v1.3 add more Usage text
 * 08/02/02 Andy Cress v1.4 moved common ipmi_cmd() code to ipmicmd.c 
 * 10/07/02 Andy Cress v1.5 added -v option with BMC version too
 * 01/16/03 Andy Cress v1.6 Handle new OS crit stop format with die code
 * 01/29/03 Andy Cress v1.7 added MV OpenIPMI support
 * 02/05/03 Andy Cress v1.8 show better message if empty SEL (cc=0xCB),
 *                          show an additional warning if free space is low.
 * 02/18/03 Andy Cress v1.9 trim out some fields so it fits on 1 line,
 *                          decode Boot Event subcodes
 * 02/27/03 Andy Cress v1.10 change OS Crit Stop decoding to handle new types
 * 03/20/03 Andy Cress v1.11 for -w, save id also, so we don't have to start
 *                           over at the beginning each time.
 * 04/30/03 Andy Cress v1.12 changed display ordering.
 * 06/23/03 Andy Cress v1.13 fix -w if log gets cleared
 * 08/19/03 Andy Cress v1.14 handle OEM/other record types
 * 09/16/03 Andy Cress v1.15 added more sens_desc strings
 * 10/03/03 Andy Cress v1.16 added more sens_desc strings (for boot)
 * 01/19/04 Andy Cress v1.17 added more sens_desc for Fans
 * 01/30/04 Andy Cress v1.18 added WIN32 flags, and sens_desc for Voltage
 * 03/12/04 Andy Cress v1.19 ClockSync description changed
 * 03/29/04 Andy Cress v1.20 change pattern matching for thresholds, 
 *                           added sens_desc for ID Button
 *                           check either time or record_id if fwritesel
 * 			     show warning if <20% free also
 *                           added sens_desc for System Events, HSC, Power, Int
 * 04/13/04 Andy Cress v1.21 added threshold OK descriptions,
 *                           change header (time is local, not GMT)
 * 05/05/04 Andy Cress v1.22 call ipmi_close before exit,
 *                           include ReportEvent code for WIN32
 * 06/10/04 Andy Cress v1.23 use gmtime instead of localtime for WIN32
 * 08/16/04 Andy Cress v1.24 added more decoding for Power events
 * 09/20/04 Andy Cress v1.25 added 2 event descriptors for ia64 platforms
 * 11/01/04 Andy Cress v1.26 add -N / -R for remote nodes   
 * 11/16/04 Andy Cress v1.27 add -U for username,
 *                           added more decoding for mBMC watchdog events
 * 11/19/04 Andy Cress v1.28 added more decoding for crit_int, slots, etc.
 *                           changed firmware error decoding.
 * 02/17/05 Andy Cress v1.29 made decode_sel_entry() a subroutine, 
 *                           added logic for OEM 0xc0 record types.
 * 05/24/05 Andy Cress v1.30 fixed SegFault with StartWriting/fscanf
 * 05/26/05 Andy Cress v1.31 moved decode_sel_entry to events.c
 * 08/01/05 Andy Cress v1.32 updated events.c for PowerUnit & Battery
 * 09/12/05 Andy Cress v1.33 dont check superuser for fipmi_lan
 * 06/29/06 Andy Cress v1.34 added -l option
 */
/*M*
Copyright (c) 2002-2005, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#include "getopt.h"
#else
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <getopt.h>
#endif
#include <time.h>

#include "ipmicmd.h" 

#define  SELprintf  printf
 
#pragma pack(1)
typedef struct
{
        ushort             record_id;
        uchar              record_type;
        uint               timestamp;
        ushort             generator_id;    /*slave_addr/channel*/
        uchar              evm_rev;         //event message revision
        uchar              sensor_type;
        uchar              sensor_number;
        uchar              event_trigger;
        uchar              event_data1;
        uchar              event_data2;
        uchar              event_data3;
}       SEL_RECORD;
#pragma pack()

#define MIN_FREE       128    /*=8*16, minimal bytes of free SEL space */
#define REC_SIZE        16    /*SEL Record Size, see IPMI 1.5 Table 26-1 */
#define RECORD_BASE     2     //base value to the SEL record in IMB resp data
#define RID_OFFSET      0     //byte offset to the record id
#define RTYPE_OFFSET    2     //byte offset to the record type
#define RTS_OFFSET      3     //byte offset to the record timestamp
#define RGID_OFFSET     7     //byte offset to the record generator id
#define REREV_OFFSET    9     //byte offset to the record event message rev
#define RSTYPE_OFFSET   10    //byte offset to the record sensor type
#define RSN_OFFSET      11    //byte offset to the record sensor number
#define RET_OFFSET      12    //byte offset to the record event trigger
#define RDATA_OFFSET    13    //byte offset to the record event data  

static char *progname  = "showsel";
static char *progver   = "1.47";
#ifdef WIN32
static char idxfile[] = "sel.idx";  /* Should this be C:\sel.idx ? */
static HANDLE hLog = NULL;
#define SELMSG_ID  0x40000101   /* showselmsg.dll EventID 257. = 0x101 */
#else
static char idxfile[] = "/usr/share/ipmiutil/sel.idx";
#endif
static char fdebug = 0;
static char fall = 1;
static char fwritesel = 0;
static char fclearsel = 0;
static char fonlyver = 0;
static char flastrecs = 0;
static uint savtime = 0;
static short savid = 0;
static int nlast = 20;
static ushort idinc = REC_SIZE;

/*------------------------------------------------------------------------ 
 * decode_sel_entry
 * Parse and decode the SEL record into readable format.
 * See events.c
 *------------------------------------------------------------------------*/
extern void decode_sel_entry( SEL_RECORD *psel, char *outbuf);

/*F*
////////////////////////////////////////////////////////////////////////////////
//  GetSelEntry
////////////////////////////////////////////////////////////////////////////////
//  Name			:	GetSelEntry
//
//  Purpose			:	This routine gets the next SEL record
//						
//  Parameters		:	
//	pRecordID		:	input/output pointer to next Record ID
//	selRecord		:	output pointer to the sel record
//
//  Returns			:	0 if success
//					-1 if last record.
//					-2 if completion code error.
//					-3 if null buffer on input.
//					-4 if record id mismatch.
//						
//	Notes			:	uses ipmi_cmd()
//
*F*/
int GetSelEntry(ushort *pRecordID, SEL_RECORD *selRecord)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar inputData[6];
	uchar completionCode;
	char* cpRecordID;	 //for next record id
	ushort inRecordID;    //save the input record id

	inRecordID = *pRecordID;
	if (!selRecord) 
	{
		if (fdebug)
		    SELprintf("GetSelEntry: error, output buffer is NULL\n");
		return -3;
	}

	//set the reservation id to zero
	inputData[0]            = 0;
	inputData[1]            = 0;

	//set the record id to get
	cpRecordID              = (char*) pRecordID;
	inputData[2]            = cpRecordID[0];
	inputData[3]            = cpRecordID[1];

	//set the offset to the record to value zero
	inputData[4]            = 0;

	//set the number of byte to read.  0xFF means read the entire record
	inputData[5]            = (char) 0xFF;

	status = ipmi_cmd(GET_SEL_ENTRY, inputData, 6, responseData, 
			&responseLength, &completionCode, fdebug);

	if (status == ACCESS_OK) {
		if( completionCode ) {
			if (completionCode == 0xCB && inRecordID == 0)
			  SELprintf("Firmware Log (SEL) is empty\n");
			else 
			  SELprintf("GetSelEntry: completion code=%x\n", 
				completionCode); // responseData[0]);
		} else {
			//successful, done

			//save the next SEL record id
			cpRecordID[0] = responseData[0];
			cpRecordID[1] = responseData[1];

			//save the SEL record content
			//(note that selRecord structure must be pragma_pack'd)
			memcpy(selRecord,&responseData[RECORD_BASE],16);

			if (inRecordID == 0 || inRecordID == selRecord->record_id)
			{
				/* We return success if the input record is 
				   begin-of-SEL (value 0), or             
                                   input and output record id matches.   
				*/
				return 0;
			}
			else 
			{
				/* If last record, inRecordID will be -1 from
				   response data, so return -1 as normal EOF. 
				   (fix to bug 279)
				*/
				if (inRecordID == 0xFFFF) return(-1);
				/* If not last record, this is an error. */
				if (fdebug)
				SELprintf("GetSelEntry: input id  %d != output id %d \n", inRecordID, selRecord->record_id);
				*pRecordID = inRecordID;  //restore the input record id
				return -4;
			}
		}
	}

	// we are here because after the retry, completionCode is not COMPLETION_CODE_OK

	if (fdebug) 
		SELprintf("GetSelEntry: sendImbRequest error %d completion code, code=%d\n", status,completionCode);
 	return -2;
 
}  /* end GetSelEntry() */

void StartWriting(uint *plasttime, short *plastid)
{
	FILE *fd;
	uint lasttime; 
	uint lastid;
	int ret = -1;
	
	lasttime = 0;
	lastid = 0;
	// Open the index file
	fd = fopen(idxfile,"r");
	if (fd != NULL) {
		// Read the file, get savtime & savid
		ret = fscanf(fd,"%x %x",&lasttime,&lastid);
		fclose(fd);
	}
	else printf("StartWriting: cannot open %s\n",idxfile);
	if (fdebug) printf("StartWriting: idx fd=%x, savtime=%x, savid=%x\n",
				fd,lasttime,(ushort)lastid);
	*plasttime = lasttime;
	*plastid = (short)lastid;

#ifdef WIN32
	/* See Registry, under
	   HKLM/SYSTEM/CurrentControlSet/Services/EventLog/Application/showsel 
	   EventMessageFile REG_EXPAND_SZ "%SystemRoot%\system32\showselmsg.dll"
	   TypesSupported   REG_DWORD     0x000000007 
	*/
	hLog = RegisterEventSource(NULL, "showsel");
	if (hLog == (void *)ERROR_INVALID_HANDLE) { hLog = NULL; }
	if (hLog == NULL) 
		printf("RegisterEventSource error, %lx\n", GetLastError());
	else ret = 0; /*success*/
#else
	// Open syslog
        openlog( "SEL: ",  LOG_CONS, LOG_KERN);
	ret = 0;   /* success if here */
#endif
	if (fdebug) printf("StartWriting: ret = %d\n",ret);
	return;
}

void StopWriting(uint lasttime, short lastid)
{
	FILE *fd;
	// Rewrite the saved time & record id
	fd = fopen(idxfile,"w");
	if (fd != NULL) {
		fprintf(fd,"%x %x\n",lasttime,lastid);
		fclose(fd);
		}
	else printf("StopWriting: cannot open %s\n",idxfile);
#ifdef WIN32
	DeregisterEventSource(hLog);
#else
	// Close syslog
	closelog();
#endif
	return;
}

void WriteSyslog(char *msgbuf)
{
#ifdef WIN32
	boolean status;
	char *rgstrings[2] = {NULL, NULL};
	rgstrings[0] = msgbuf; /*decoded SEL entry*/
	if (hLog != NULL) {
		status = ReportEvent(hLog,EVENTLOG_INFORMATION_TYPE,
					0, SELMSG_ID, NULL, 1,0,
					rgstrings,NULL);
					/* showsel eventid = 0x101. */
		if (fdebug || (status == 0)) {  /*error or debug*/
		      printf("ReportEvent status=%d, %lx\n",
					status,GetLastError());
		} 
	}
#else
	syslog(LOG_INFO,"%s",msgbuf);
#endif
} /*end WriteSyslog*/

void ClearSEL(void)
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar inputData[6];
	uchar completionCode;
	ushort cmd;

	cmd = RESERVE_SEL;
	status = ipmi_cmd(cmd, inputData, 0, responseData, 
			&responseLength, &completionCode, fdebug);

	if (status == ACCESS_OK && completionCode == 0) {
	    cmd = CLEAR_SEL;
	    inputData[0] = responseData[0];  /*Reservation ID [0]*/
	    inputData[1] = responseData[1];  /*Reservation ID [1]*/
	    inputData[2] = 'C';
	    inputData[3] = 'L';
	    inputData[4] = 'R';
	    inputData[5] = 0xAA;   /* initiate erase */
	    status = ipmi_cmd(cmd, inputData, 6, responseData, 
	  		      &responseLength, &completionCode, fdebug);
	    /* The reservation is cancelled by the CLEAR_SEL cmd */
	    }
	if (status == ACCESS_OK) {
		if( completionCode ) {
		    SELprintf("ClearSEL: cmd = %x, completion code=%x\n",
			      cmd,completionCode); 
		} else {
		    /* Successful, done. */
		    SELprintf("ClearSEL: Log Cleared successfully\n"); 
		}
	} else {
	    SELprintf("ClearSEL: cmd = %x, error status = %x\n",
		      cmd,status); 
	}
	if (fwritesel) {  /* this won't always be enough, but try */
	   StartWriting(&savtime,&savid);
	   savid = 0;  
	   StopWriting(savtime,savid);
	}
}  /* ClearSEL()*/


void ReadSEL(uchar mytype, char fwriteit)
{
	ushort RecordID = 0;  /* 0 = first record, 0xFFFF = end */
	SEL_RECORD selRecord;
	SEL_RECORD *pSelRecord = &selRecord;
	char output[132];
	int rc = 0;
        int ilast = 0;
        short recid0;

	if (fwriteit) { 
		StartWriting(&savtime,&savid);
		RecordID = savid;
	}
	memset(pSelRecord, 0, sizeof(SEL_RECORD));
        if (flastrecs) { 
            /* read the first two records to get the record id increment */
            RecordID = 0;
	    rc = GetSelEntry( &RecordID, pSelRecord);
	    recid0 = pSelRecord->record_id;
	    rc = GetSelEntry( &RecordID, pSelRecord);
            idinc = pSelRecord->record_id - recid0;
            RecordID = 0xFFFF;  /* -1=0xFFFF; * get last record */
            if (fdebug) printf("recid inc = 0x%02x\n",idinc);
        }
	/* show header for the SEL records */
        printf("RecId Date/Time_______ Source_ Evt_Type SensNum Evt_detail - Trig [Evt_data]\n");
	while( rc == 0 ) {
		rc = GetSelEntry( &RecordID, pSelRecord);
		if (fwriteit && (rc != 0) && (RecordID == savid)) {
		    /* If here, log was probably cleared, so try
		     * again from the log start. */
		    RecordID = 0;
		    rc = GetSelEntry( &RecordID, pSelRecord);
		}
                if (fdebug) printf("rc = %d, recid = %04x, next = %04x\n",
                                   rc, pSelRecord->record_id, RecordID);
                if (flastrecs && (ilast == 0) && (rc == -1)) rc = 0;
		if (rc != 0) { /* EOF or error */ break; }

		if (mytype == 0xff || pSelRecord->sensor_type == mytype) {
		   decode_sel_entry(pSelRecord,output);
		   printf("%s",output);
		} else {
		   if (fdebug) printf("error, mytype = %d\n",mytype);
		   output[0] = 0;
		}
	
		if (fwriteit) {
		   /* Only write newer records  to syslog */
		   if (pSelRecord->record_type == 0x02) {
			if ((pSelRecord->timestamp > savtime) ||
			    (pSelRecord->record_id > savid)) {
			   WriteSyslog(output);
			   savid = pSelRecord->record_id;
			   savtime = pSelRecord->timestamp;
			}
		   } else {   /* no timestamp */
			if (pSelRecord->record_id > savid) {
			   WriteSyslog(output);
			   savid = pSelRecord->record_id;
			}
		   }
		}  /*endif writeit*/
		if( pSelRecord->record_id == 0xFFFF )
			break;
		if( RecordID == pSelRecord->record_id )
			break;
                if (flastrecs) { 
                    ++ilast;
                    if (ilast >= nlast) break;
                    RecordID = pSelRecord->record_id - idinc;
                    if (RecordID > pSelRecord->record_id) break;
                }
		memset(pSelRecord, 0, sizeof(SEL_RECORD));
	}  /*endwhile*/
	if (fwriteit) StopWriting(savtime,savid);
}  /* end ReadSEL()*/

ushort vfree = 0;
ushort vtotal = 0;

static int ReadSELinfo()
{
	uchar responseData[MAX_BUFFER_SIZE];
	int responseLength = MAX_BUFFER_SIZE;
	int status;
	uchar completionCode;
	uchar inputData[6];

	status = ipmi_cmd(GET_SEL_INFO, inputData, 0, responseData, 
			&responseLength, &completionCode, fdebug);
	if (fdebug) printf("ipmi_cmd status = %x\n",status);

	if (status == ACCESS_OK) {
		vfree = responseData[3] + (responseData[4] << 8);
		vtotal = responseData[1] + (responseData[2] << 8) + (vfree/REC_SIZE);
		SELprintf("SEL Ver %x Support %x, Size = %d records, Free space = %d records\n",
		          responseData[0],
		          responseData[13],
			  vtotal,
			  vfree/REC_SIZE );
		//successful, done
		return(0);
	} else    return(1);

}  /*end ReadSELinfo()*/


#ifdef METACOMMAND
int i_sel(int argc, char **argv)
#else
#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
#endif
{
   int ret;
   char DevRecord[16];
   char c;
   // int i;

   printf("%s: version %s\n",progname,progver);
   while ( (c = getopt( argc, argv,"cswvxl:T:V:EYF:P:N:U:R:?")) != EOF )
      switch(c) {
          case 'x': fdebug = 1;      break;
          case 'w': fwritesel = 1;   break;
          case 'l': flastrecs = 1; 
                nlast = atoi(optarg);
                break;
          case 's': fall = 0;        break;
          case 'c': fclearsel = 1;   break;
          case 'v': fonlyver = 1;    break;
          case 'N':    /* nodename */
          case 'U':    /* remote username */
          case 'P':    /* remote password */
          case 'R':    /* remote password */
          case 'E':    /* get password from IPMI_PASSWORD environment var */
          case 'F':    /* force driver type */
          case 'T':    /* auth type */
          case 'V':    /* priv level */
          case 'Y':    /* prompt for remote password */
                parse_lan_options(c,optarg,fdebug);
                break;
          default:
                printf("Usage: %s [-cswvx] [-l 5] [-NUPREFTVY]\n",
                       progname);
		printf("   -c  Clears all SEL records\n");
		printf("   -s  Show only Critical Stop records\n");
		printf("   -l5 Show last 5 SEL records (reverse order)\n");
		printf("   -w  Writes new SEL records to syslog\n");
		printf("   -v  Only show version information\n");
		printf("   -x  Display extra debug messages\n");
		print_lan_opt_usage();
                exit(1);
      }               
   if (fwritesel && flastrecs) {
       printf("Error: Options -l and -w are incompatible\n");
       exit(1);
   }
   // if (fwrite) fall = 1;   // would we ever set fwrite without fall??
#ifndef WIN32
    if (fipmi_lan == 0) {
        /* only run this as superuser */
        ret = geteuid();
        if (ret > 1) {
            printf("Not superuser (%d)\n", ret);
            exit(1);
        }
    }
#endif

   ret = ipmi_getdeviceid( DevRecord, sizeof(DevRecord),fdebug);
   if (ret == 0) {
      uchar ipmi_maj, ipmi_min;
      ipmi_maj = DevRecord[4] & 0x0f;
      ipmi_min = DevRecord[4] >> 4;
      printf("-- BMC version %x.%x, IPMI version %d.%d \n",
             DevRecord[2],  DevRecord[3],
             ipmi_maj, ipmi_min);
   } else {
      printf("GetDeviceID ret = %d.  IPMI session failure.\n",ret);
      ipmi_close_();
      exit(ret);
   }

   ret = ReadSELinfo();
   if (ret == 0 && !fonlyver) {
      if (fclearsel) {
	 ClearSEL();
      } else {
	 if (fdebug) printf("%s: starting ReadSEL ...\n",progname);
         if (fall) ReadSEL(0xff,fwritesel); /* show all SEL records */
         else ReadSEL(0x20,fwritesel);      /* only show OS Crit Stops*/
	 /* PEF alerts and other log messages fail if low free space,
	    so show a warning. */
	 if (vfree < MIN_FREE) {
             printf("WARNING: free space is very low (=%d), need to clear with -c\n",
		    vfree);
	 } else if ((vfree/REC_SIZE) < ((vtotal * 20)/100)) {
             printf("WARNING: free space is low (=%d), need to clear with -c\n",
		    vfree);
	 }
      }
   }
   ipmi_close_();
   exit(ret);
}

/* end showsel.c */
